/* Copyright (C) |Meso|Star> 2015-2017 (contact@meso-star.com)
 *
 * This software is a library whose purpose is to generate [pseudo] random
 * numbers and random variates.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp_rng_c.h"

#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>

using piecewise_dist = RAN_NAMESPACE::piecewise_linear_distribution<double>;

struct ssp_ranst_piecewise_linear {
  ref_T ref;
  struct mem_allocator* allocator;
  piecewise_dist* distrib;
};

/*******************************************************************************
 * Helper function
 ******************************************************************************/
static void
piecewise_release(ref_T* ref)
{
  ssp_ranst_piecewise_linear* ran;
  ASSERT(ref);
  ran = CONTAINER_OF(ref, ssp_ranst_piecewise_linear, ref);
  if(ran->distrib) {
    ran->distrib->~piecewise_dist();
    MEM_RM(ran->allocator, ran->distrib);
  }
  MEM_RM(ran->allocator, ran);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
ssp_ranst_piecewise_linear_create
  (struct mem_allocator* allocator,
   struct ssp_ranst_piecewise_linear** out_ran)
{
  struct ssp_ranst_piecewise_linear* ran = nullptr;
  void* mem = nullptr;
  res_T res = RES_OK;

  if(!out_ran) {
    res = RES_BAD_ARG;
    goto error;
  }

  allocator = allocator ? allocator : &mem_default_allocator;

  ran = static_cast<ssp_ranst_piecewise_linear*>
    (MEM_CALLOC(allocator, 1, sizeof(ssp_ranst_piecewise_linear)));
  if(!ran) {
    res = RES_MEM_ERR;
    goto error;
  }
  ref_init(&ran->ref);

  mem = MEM_ALLOC(allocator, sizeof(piecewise_dist));
  if(!mem) {
    res = RES_MEM_ERR;
    goto error;
  }
  ran->allocator = allocator;
  ran->distrib = static_cast<piecewise_dist*>(new(mem) piecewise_dist);
  if(out_ran) *out_ran = ran;

exit:
  return res;
error:
  if(ran) {
    SSP(ranst_piecewise_linear_ref_put(ran));
    ran = nullptr;
  }
  goto exit;
}

res_T
ssp_ranst_piecewise_linear_setup
  (struct ssp_ranst_piecewise_linear* ran,
   const double* intervals,
   const double* weights,
   const size_t size)
{
  if(!ran || !intervals || !weights || size < 2)
    return RES_BAD_ARG;

  piecewise_dist::param_type p{intervals, intervals + size, weights};
  ran->distrib->param(p);
  return RES_OK;
}

res_T
ssp_ranst_piecewise_linear_ref_get
  (struct ssp_ranst_piecewise_linear* ran)
{
  if(!ran) return RES_BAD_ARG;
  ref_get(&ran->ref);
  return RES_OK;
}

res_T
ssp_ranst_piecewise_linear_ref_put
  (struct ssp_ranst_piecewise_linear* ran)
{
  if(!ran) return RES_BAD_ARG;
  ref_put(&ran->ref, piecewise_release);
  return RES_OK;
}

double
ssp_ranst_piecewise_linear_get
  (const struct ssp_ranst_piecewise_linear* ran,
   struct ssp_rng* rng)
{
  class rng_cxx r(*rng);
  return ran->distrib->operator()(r);
}

